from typing import List


class CNode(object):
    _children: List['CNode']

    def __init__(
            self,
            *,
            _id: int,
            callee_region_id: int
    ):
        self._children = []
        self.id = _id
        self.callee_region_id = callee_region_id

    def get_children(self):
        return self._children

    def add_child(self, child: 'CNode'):
        self._children.append(child)

    def get_all_children(self) -> List['CNode']:
        cnodes = [self]
        for child in self._children:
            cnodes += child.get_all_children()
        return cnodes

    def __repr__(self):
        return 'CNode<{}>'.format(self.__dict__)



import struct
from typing import List, BinaryIO, Any

from pycubexr.utils.metric_formats import METRIC_FORMATS

DATA_HEADER: bytes = b'CUBEX.DATA'


def parse_data(
        *,
        data_file: BinaryIO,
        # From the anchor.xml. NOT the Python struct data types, e.g., "i"
        data_type: str,
        # Either "<" or ">"
        endianness_format_char: str
) -> List[Any]:
    # Verify the data file header
    header = data_file.read(len(DATA_HEADER))
    assert header == DATA_HEADER

    raw = data_file.read()

    # Calculate the size for a single element
    single_value_size = struct.calcsize(METRIC_FORMATS[data_type])

    # Verify that the number of read bytes is divisible by the value size
    assert len(raw) % single_value_size == 0

    num_values = int(len(raw) / single_value_size)

    # Example: the format '<100i' means to parse/"unpack" 100 integers
    unpack_format = endianness_format_char + str(num_values) + METRIC_FORMATS[data_type]

    return list(struct.unpack(unpack_format, raw))



import tarfile
from tarfile import TarFile
from typing import List, Dict
from xml.etree import ElementTree

from pycubexr.classes import Metric, MetricValues, Region, CNode, Location
from pycubexr.parsers.anchor_xml_parser import parse_anchor_xml, AnchorXMLParseResult
from pycubexr.parsers.metrics_parser import extract_metric_values
from pycubexr.utils.exceptions import MissingMetricError


class CubexParser(object):
    _cubex_file: TarFile
    _cubex_filename: str
    _anchor_result: AnchorXMLParseResult
    _metric_values: Dict[int, MetricValues]

    def __init__(self, cubex_filename: str):
        self._cubex_filename = cubex_filename
        self._metric_values = {}

    def __enter__(self):
        self._cubex_file = tarfile.open(self._cubex_filename)

        with self._cubex_file.extractfile('anchor.xml') as anchor_file:
            anchor = ElementTree.parse(anchor_file)
            self._anchor_result = parse_anchor_xml(anchor)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cubex_file.close()

    def get_metric_values(
            self,
            metric: Metric
    ) -> MetricValues:
        if metric.id in self._metric_values:
            return self._metric_values[metric.id]

        index_file_name = '{}.index'.format(metric.id)
        data_file_name = '{}.data'.format(metric.id)

        if index_file_name not in self._tar_file_members():
            raise MissingMetricError(metric)

        with self._cubex_file.extractfile(index_file_name) as index_file, \
                self._cubex_file.extractfile(data_file_name) as data_file:
            metric_values = extract_metric_values(
                metric=metric,
                index_file=index_file,
                data_file=data_file
            )

            assert metric_values.num_locations() == len(self.get_locations())

            self._metric_values[metric.id] = metric_values
            return metric_values

    def get_metrics(self):
        return self._anchor_result.metrics

    def get_metric_by_name(self, metric_name: str) -> Metric:
        return [metric for metric in self._anchor_result.metrics if metric.name == metric_name][0]

    def get_region(self, cnode: CNode) -> Region:
        return [region for region in self._anchor_result.regions if region.id == cnode.callee_region_id][0]

    def get_cnode(self, cnode_id: int) -> CNode:
        return [cnode for cnode in self._anchor_result.cnodes[0].get_all_children() if cnode.id == cnode_id][0]

    def get_locations(self) -> List[Location]:
        return self._anchor_result.system_tree_nodes[0].all_locations()
    
    def get_calltree(self, indent=0, cnode: CNode = None):
        if cnode is None:
            cnode = self._anchor_result.cnodes[0]
        call_tree_string = ""
        child_string = ""
        child_string += "-"*indent + self.get_region(cnode).name
        call_tree_string += child_string
        call_tree_string += "\n"
        
        for child in cnode.get_children():
            tmp = self.get_calltree(indent + 1, cnode=child)
            if tmp is not None:
                call_tree_string += tmp

        return call_tree_string

    def print_calltree(self, indent=0, cnode: CNode = None):
        if cnode is None:
            cnode = self._anchor_result.cnodes[0]

        print('\t' * indent, self.get_region(cnode).name)

        for child in cnode.get_children():
            self.print_calltree(indent + 1, cnode=child)

    def _tar_file_members(self) -> List[str]:
        return [x.name for x in self._cubex_file.getmembers()]



from pycubexr.classes.cnode import CNode
from pycubexr.classes.location import Location
from pycubexr.classes.location_group import LocationGroup
from pycubexr.classes.metric import Metric
from pycubexr.classes.metric_values import MetricValues
from pycubexr.classes.region import Region
from pycubexr.classes.system_tree_node import SystemTreeNode



import struct
from typing import List, BinaryIO, Any

from pycubexr.utils.exceptions import CorruptIndexError

INDEX_HEADER: bytes = b'CUBEX.INDEX'


class IndexParseResult(object):

    def __init__(self, endianness_format: str, cnode_indices: List[Any]):
        self.endianness_format = endianness_format
        self.cnode_indices = cnode_indices


def parse_index(index_file: BinaryIO) -> IndexParseResult:
    assert not index_file.closed

    header = index_file.read(len(INDEX_HEADER))
    assert header == INDEX_HEADER

    # The index file starts with the number 1 encoded as an int.
    # Here, we parse it to determine the endianness of the machine that created the index file.
    one = struct.unpack('<i', index_file.read(4))[0]

    # If the "one" is parsed as a 1, the endianness "<" format is right - else flip it.
    # Important: the endianness must also be used for the data file!
    endianness_format = '<' if one == 1 else '>'

    # IMPORTANT: not used but MUST not be deleted!
    # noinspection PyUnusedLocal
    version = struct.unpack('{}h'.format(endianness_format), index_file.read(2))[0]
    # noinspection PyUnusedLocal
    index_type = index_file.read(1)

    # Parse how many indices this file contains
    raw_size = index_file.read(4)
    n_nodes = struct.unpack('{}i'.format(endianness_format), raw_size)[0]

    raw_index = index_file.read()

    index_value_size = struct.calcsize("i")
    if len(raw_index) != index_value_size * n_nodes:
        raise CorruptIndexError(
            "The size of the index list should be equal to (size of one index value * number of nodes)."
        )

    cnode_indices = list(struct.unpack('{}{}i'.format(endianness_format, n_nodes), raw_index))
    assert len(cnode_indices) > 0
    assert index_file.peek() == b''
    return IndexParseResult(
        endianness_format=endianness_format,
        cnode_indices=cnode_indices
    )



from typing import Dict, List
from xml.etree import ElementTree

from pycubexr.classes import Metric, Region, CNode, SystemTreeNode
from pycubexr.parsers import xml_parser_helper


class AnchorXMLParseResult(object):

    def __init__(
            self,
            *,
            attrs: Dict[str, str],
            metrics: List[Metric],
            regions: List[Region],
            cnodes: List[CNode],
            system_tree_nodes: List[SystemTreeNode]
    ):
        self.attrs = attrs
        self.metrics = metrics
        self.regions = regions
        self.cnodes = cnodes
        self.system_tree_nodes = system_tree_nodes


def parse_anchor_xml(root: ElementTree):
    result = AnchorXMLParseResult(
        attrs=xml_parser_helper.parse_attrs(root),
        metrics=xml_parser_helper.parse_metrics(root),
        regions=xml_parser_helper.parse_regions(root),
        cnodes=xml_parser_helper.parse_cnodes(root),
        system_tree_nodes=xml_parser_helper.parse_system_tree_nodes(root),
    )
    assert len(result.cnodes) == 1
    assert len(result.system_tree_nodes) == 1
    return result



import typing
from typing import List

from pycubexr.classes import LocationGroup, Location


class SystemTreeNode(object):
    _system_tree_node_children: List['SystemTreeNode']
    _location_group_children: List[LocationGroup]

    def __init__(
            self,
            *,
            _id: int,
            _class: str,
            name: str,
            attrs: typing.Dict[str, str]
    ):
        self._system_tree_node_children = []
        self._location_group_children = []
        self.id = _id
        self.clazz = _class
        self.name = name
        self.attrs = attrs

    def all_location_groups(self) -> List[LocationGroup]:
        location_groups = [x for x in self._location_group_children]
        for system_tree_child in self._system_tree_node_children:
            location_groups += system_tree_child.all_location_groups()
        return location_groups

    def all_locations(self) -> List[Location]:
        locations: List[Location] = []
        for location_group in self.all_location_groups():
            locations += location_group.all_locations()
        return locations

    def add_system_tree_node_child(self, child: 'SystemTreeNode'):
        self._system_tree_node_children.append(child)

    def add_location_group(self, child: LocationGroup):
        self._location_group_children.append(child)

    def __repr__(self):
        return 'SystemTreeNode<{}>'.format(self.__dict__)



from pycubexr.utils.metric_formats import METRIC_FORMATS


class MetricType(object):
    INCLUSIVE = 'INCLUSIVE'
    EXCLUSIVE = 'EXCLUSIVE'


class Metric(object):

    def __init__(
            self,
            *,
            name: str,
            _id: int,
            display_name: str,
            description: str,
            metric_type: str,
            data_type: str,
            units: str,
            url: str
    ):
        assert hasattr(MetricType, metric_type)
        assert data_type in METRIC_FORMATS
        self.name = name
        self.id = _id
        self.display_name = display_name
        self.description = description
        self.metric_type = metric_type
        self.data_type = data_type
        self.units = units
        self.url = url

    def __repr__(self):
        return 'Metric<{}>'.format(self.__dict__)



from typing import List, Any


class MetricValues(object):

    def __init__(
            self,
            *,
            cnode_indices: List[int],
            values: List[Any]
    ):
        self.values = values
        self.cnode_indices = cnode_indices
        assert len(self.values) % len(self.cnode_indices) == 0

    def num_locations(self):
        return int(len(self.values) / len(self.cnode_indices))

    def cnode_values(self, cnode_id: int):
        assert cnode_id in self.cnode_indices
        start_index = int(self.cnode_indices.index(cnode_id) * self.num_locations())
        end_index = start_index + self.num_locations()
        return self.values[start_index:end_index]

    def location_value(self, cnode_id: int, location_id: int):
        assert location_id < self.num_locations()
        return self.cnode_values(cnode_id)[location_id]

    def __repr__(self):
        return 'MetricValues<{}>'.format(self.__dict__)



import setuptools

with open("README.md") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pycubexr",
    version="0.0.1",
    author="Extra-P project",
    author_email="extra-p@lists.parallel.informatik.tu-darmstadt.de",
    description="pyCubeR is a Python package for reading the Cube4 file format.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/extra-p/pycubexr",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)
